// Copyright (C) 2016 The Android Open Source Project
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto2";

option optimize_for = LITE_RUNTIME;

message DataContainer {
    repeated AppData app = 1;
    repeated CpuData cpu = 2;
};

message CpuData {
    required int64 timestamp = 1;
    required int64 uptime = 2;
    required int64 utime = 3;
    required int64 ntime = 4;
    required int64 stime = 5;
    required int64 itime = 6;
    required int64 iowtime = 7;
    required int64 irqtime = 9;
    required int64 sirqtime = 10;
}

message AppData {
    required int32 pid = 1;
    required string tname = 2;
    required string name = 3;
    repeated AppSample samples = 4;
}

message AppSample {
    required int64 timestamp = 1;
    required int64 uptime = 2;
    required int64 rchar = 3;
    required int64 wchar = 4;
    required int64 syscr = 5;
    required int64 syscw = 6;
    required int64 readbytes = 7;
    required int64 writebytes = 8;
    required int64 utime = 9;
    required int64 stime = 10;
    required int64 rss = 11;
};
