# This file is the LOCAL_INIT_RC file for the bootio command.

service bootio /system/bin/bootio
    class main
    user root
    group root
    oneshot

on post-fs-data
    mkdir /data/misc/bootio 0755 root shell
