/*
 * Copyright (C) 2016 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <unistd.h>

#include <gtest/gtest.h>

TEST(kernel_config, NOT_CONFIG_ANDROID_LOGGER) {
  EXPECT_NE(0, access("/dev/log/main", F_OK));
  EXPECT_NE(0, access("/dev/log_main", F_OK));
  EXPECT_NE(0, access("/dev/log/radio", F_OK));
  EXPECT_NE(0, access("/dev/log_radio", F_OK));
  EXPECT_NE(0, access("/dev/log/events", F_OK));
  EXPECT_NE(0, access("/dev/log_events", F_OK));
  EXPECT_NE(0, access("/dev/log/system", F_OK));
  EXPECT_NE(0, access("/dev/log_system", F_OK));
}
