// Copyright (C) 2015 The Android Open Source Project
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <map>
#include <vector>

#ifndef _IOTOP_TASKLIST_H
#define _IOTOP_TASKLIST_H

class TaskList {
public:
  static bool Scan(std::map<pid_t, std::vector<pid_t>>&);

private:
  TaskList() {}
  static bool ScanPid(pid_t pid, std::vector<pid_t>&);
};

#endif // _IOTOP_TASKLIST_H
